@@ -13,13 +13,15 @@
 //! roughly twice as slow. Random indexing has identical performance for both.
 
 use std;
+#[cfg(feature = "unstable")]
+use std::collections::TryReserveError;
 use std::iter::FusedIterator;
 use std::marker::PhantomData;
 use std::mem::ManuallyDrop;
 use std::ops::{Index, IndexMut};
 use std::{fmt, ptr};
 
-use super::{DefaultKey, Key, KeyData, Slottable};
+use crate::{DefaultKey, Key, KeyData, Slottable};
 
 // Metadata to maintain the freelist.
 #[derive(Clone, Copy, Debug)]
@@ -48,7 +50,13 @@ enum SlotContent<'a, T: 'a + Slottable> {
     Vacant(&'a FreeListEntry),
 }
 
+enum SlotContentMut<'a, T: 'a + Slottable> {
+    OccupiedMut(&'a mut T),
+    VacantMut(&'a mut FreeListEntry),
+}
+
 use self::SlotContent::{Occupied, Vacant};
+use self::SlotContentMut::{OccupiedMut, VacantMut};
 
 impl<T: Slottable> Slot<T> {
     // Is this slot occupied?
@@ -66,6 +74,16 @@ impl<T: Slottable> Slot<T> {
             }
         }
     }
+
+    pub fn get_mut(&mut self) -> SlotContentMut<T> {
+        unsafe {
+            if self.occupied() {
+                OccupiedMut(&mut *self.u.value)
+            } else {
+                VacantMut(&mut self.u.free)
+            }
+        }
+    }
 }
 
 impl<T: Slottable> Drop for Slot<T> {
@@ -268,6 +286,26 @@ impl<K: Key, V: Slottable> HopSlotMap<K, V> {
         self.slots.reserve(needed);
     }
 
+    /// Tries to reserve capacity for at least `additional` more elements to be
+    /// inserted in the `HopSlotMap`. The collection may reserve more space to
+    /// avoid frequent reallocations.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # use slotmap::*;
+    /// let mut sm = HopSlotMap::new();
+    /// sm.insert("foo");
+    /// sm.try_reserve(32).unwrap();
+    /// assert!(sm.capacity() >= 33);
+    /// ```
+    #[cfg(feature = "unstable")]
+    pub fn try_reserve(&mut self, additional: usize) -> Result<(), TryReserveError> {
+        // One slot is reserved for the freelist sentinel.
+        let needed = (self.len() + additional).saturating_sub(self.slots.len() - 1);
+        self.slots.try_reserve(needed)
+    }
+
     /// Returns `true` if the slot map contains `key`.
     ///
     /// # Examples
@@ -520,29 +558,32 @@ impl<K: Key, V: Slottable> HopSlotMap<K, V> {
     where
         F: FnMut(K, &mut V) -> bool,
     {
-        let mut elems_left_to_scan = self.len();
-        let mut cur = unsafe { self.slots.get_unchecked(0).u.free.other_end as usize + 1 };
-        while elems_left_to_scan > 0 {
-            // This is safe because removing elements does not shrink slots, cur always
-            // points to an occupied slot.
-            let idx = cur;
-            let slot = unsafe { self.slots.get_unchecked_mut(cur) };
-            let version = slot.version;
-            let key = KeyData::new(cur as u32, version).into();
-            let should_remove = !f(key, unsafe { &mut *slot.u.value });
+        let len = self.slots.len();
+        let mut i = 0;
+        while i < len {
+            let should_remove = {
+                // This is safe because removing elements does not shrink slots.
+                let slot = unsafe { self.slots.get_unchecked_mut(i) };
+                let version = slot.version;
 
-            cur = match self.slots.get(cur + 1).map(|s| s.get()) {
-                Some(Occupied(_)) => cur + 1,
-                Some(Vacant(free)) => free.other_end as usize + 1,
-                None => 0,
+                match slot.get_mut() {
+                    OccupiedMut(value) => {
+                        let key = KeyData::new(i as u32, version);
+                        !f(key.into(), value)
+                    }
+                    VacantMut(free) => {
+                        i = free.other_end as usize;
+                        false
+                    }
+                }
             };
 
             if should_remove {
-                // This must happen after getting the next index.
-                unsafe { self.remove_from_slot(idx) };
+                // This is safe because we know that the slot was occupied.
+                unsafe { self.remove_from_slot(i) };
             }
 
-            elems_left_to_scan -= 1;
+            i += 1;
         }
     }
 
@@ -1169,7 +1210,7 @@ mod serialize {
             D: Deserializer<'de>,
         {
             let mut slots: Vec<Slot<V>> = Deserialize::deserialize(deserializer)?;
-            if slots.len() >= (1 << 32) - 1 {
+            if slots.len() >= u32::max_value() as usize {
                 return Err(de::Error::custom(&"too many slots"));
             }
 
