@@ -1,7 +1,7 @@
 #![deny(warnings, missing_docs, missing_debug_implementations)]
-#![doc(html_root_url = "https://docs.rs/slotmap/0.3.2")]
+#![doc(html_root_url = "https://docs.rs/slotmap/0.4.1")]
 #![crate_name = "slotmap"]
-#![cfg_attr(feature = "unstable", feature(untagged_unions))]
+#![cfg_attr(feature = "unstable", feature(untagged_unions, try_reserve))]
 
 //! # slotmap
 //!
@@ -68,8 +68,9 @@
 //! permanently unique<sup>*</sup> reference to the inserted value. Despite
 //! supporting versioning, a [`SlotMap`] is not slower than [`slab`], by
 //! internally using carefully checked unsafe code. A [`HopSlotMap`]
-//! also provides faster iteration than [`slab`] does. Additionally, at the time
-//! of writing [`slab`] does not support serialization.
+//! also provides faster iteration than [`slab`] does, and [`DenseSlotMap`] even
+//! faster still. Additionally, at the time of writing [`slab`] does not support
+//! serialization.
 //!
 //! # Performance characteristics and implementation details
 //!
@@ -88,9 +89,10 @@
 //!
 //! The memory usage for each slot in [`SlotMap`] is `4 + max(sizeof(T), 4)`
 //! rounded up to the alignment of `T`. Similarly it is `4 + max(sizeof(T), 12)`
-//! for [`HopSlotMap`].
+//! for [`HopSlotMap`]. [`DenseSlotMap`] has an overhead of 8 bytes per element
+//! and 8 bytes per slot.
 //!
-//! # Choosing `SlotMap` or `HopSlotMap`
+//! # Choosing `SlotMap`, `HopSlotMap` or `DenseSlotMap`
 //!
 //! A [`SlotMap`] can never shrink the size of its underlying storage, because
 //! for each storage slot it must remember what the latest stored version was,
@@ -104,6 +106,13 @@
 //! [`SlotMap`] a lot, choose [`HopSlotMap`]. The downside is that insertion and
 //! removal is roughly twice as slow. Random access is the same speed for both.
 //!
+//! [`DenseSlotMap`] goes even further and stores all elements on a contiguous
+//! block of memory. It uses two indirects per random access; the slots contain
+//! indices used to access the contiguous memory. This means random access is
+//! slower than both [`SlotMap`] and [`HopSlotMap`], but iteration is
+//! significantly faster. Finally, there is no trait requirement on the value
+//! type of a [`DenseSlotMap`], see [`Slottable`] for more details.
+//!
 //! # Choosing `SecondaryMap` or `SparseSecondaryMap`
 //!
 //! You want to associate extra data with objects stored in a slot map, so you
@@ -146,9 +155,10 @@
 //! [`HashMap`]: https://doc.rust-lang.org/std/collections/struct.HashMap.html
 //! [`SlotMap`]: struct.SlotMap.html
 //! [`HopSlotMap`]: hop/struct.HopSlotMap.html
+//! [`DenseSlotMap`]: dense/struct.DenseSlotMap.html
 //! [`SecondaryMap`]: secondary/struct.SecondaryMap.html
 //! [`SparseSecondaryMap`]: sparse_secondary/struct.SparseSecondaryMap.html
-//! [`HopSlotMap`]: hop/struct.HopSlotMap.html
+//! [`Slottable`]: trait.Slottable.html
 //! [`Key`]: trait.Key.html
 //! [`new_key_type!`]: macro.new_key_type.html
 //! [`serde`]: https://github.com/serde-rs/serde
@@ -156,7 +166,6 @@
 //! [`DefaultKey`]: struct.DefaultKey.html
 
 #[cfg(feature = "serde")]
-#[macro_use]
 extern crate serde;
 
 // So our macros can refer to these.
@@ -174,51 +183,66 @@ extern crate quickcheck;
 extern crate serde_json;
 
 pub(crate) mod normal;
-pub use normal::*;
+pub use crate::normal::*;
+
+pub mod dense;
+pub use crate::dense::DenseSlotMap;
 
 pub mod hop;
-pub use hop::HopSlotMap;
+pub use crate::hop::HopSlotMap;
 
 pub mod secondary;
-pub use secondary::SecondaryMap;
+pub use crate::secondary::SecondaryMap;
 
 pub mod sparse_secondary;
-pub use sparse_secondary::SparseSecondaryMap;
+pub use crate::sparse_secondary::SparseSecondaryMap;
 
+use std::fmt::{self, Debug, Formatter};
 use std::num::NonZeroU32;
 
-/// A trait for items that can go in a slot map. Due to current stable Rust
-/// restrictions a type must be [`Copy`] to be placed in a slot map. If you must
-/// store a type that is not [`Copy`] you must use nightly Rust and enable the
-/// `unstable` feature for `slotmap` by editing your `Cargo.toml`.
+/// A trait for items that can go in a [`SlotMap`] or [`HopSlotMap`]. Due to
+/// current stable Rust restrictions a type must be [`Copy`] to be placed in one
+/// of those slot maps. This restriction does not apply to [`DenseSlotMap`],
+/// [`SecondaryMap`] or [`SparseSecondaryMap`]. It also does not apply if you
+/// use nightly Rust and enable the `unstable` feature for `slotmap` by editing
+/// your `Cargo.toml`:
 ///
 /// ```text
 /// slotmap = { version = "...", features = ["unstable"] }
 /// ```
 ///
 /// This trait should already be automatically implemented for any type that is
-/// slottable. If you can't use unstable Rust and still want to store [`Copy`]
-/// data, store that as associated data in a [`SecondaryMap`].
+/// slottable.
 ///
 /// [`Copy`]: https://doc.rust-lang.org/std/marker/trait.Copy.html
 /// [`SecondaryMap`]: secondary/struct.SecondaryMap.html
+/// [`SparseSecondaryMap`]: sparse_secondary/struct.SparseSecondaryMap.html
+/// [`SlotMap`]: struct.SlotMap.html
+/// [`HopSlotMap`]: hop/struct.HopSlotMap.html
+/// [`DenseSlotMap`]: dense/struct.DenseSlotMap.html
 #[cfg(not(feature = "unstable"))]
 pub trait Slottable: Copy {}
 
-/// A trait for items that can go in a slot map. Due to current stable Rust
-/// restrictions a type must be [`Copy`] to be placed in a slot map. If you must
-/// store a type that is not [`Copy`] you must use nightly Rust and enable the
-/// `unstable` feature for `slotmap` by editing your `Cargo.toml`.
+/// A trait for items that can go in a [`SlotMap`] or [`HopSlotMap`]. Due to
+/// current stable Rust restrictions a type must be [`Copy`] to be placed in one
+/// of those slot maps. This restriction does not apply to [`DenseSlotMap`],
+/// [`SecondaryMap`] or [`SparseSecondaryMap`]. It also does not apply if you
+/// use nightly Rust and enable the `unstable` feature for `slotmap` by editing
+/// your `Cargo.toml`:
 ///
-/// ```norun
+/// ```text
 /// slotmap = { version = "...", features = ["unstable"] }
 /// ```
 ///
 /// This trait should already be automatically implemented for any type that is
-/// slottable. If you can't use unstable Rust and still want to store [`Copy`]
-/// data, store that as associated data in a [`SecondaryMap`].
+/// slottable.
 ///
 /// [`Copy`]: https://doc.rust-lang.org/std/marker/trait.Copy.html
+/// [`SecondaryMap`]: secondary/struct.SecondaryMap.html
+/// [`SparseSecondaryMap`]: sparse_secondary/struct.SparseSecondaryMap.html
+/// [`SlotMap`]: struct.SlotMap.html
+/// [`HopSlotMap`]: hop/struct.HopSlotMap.html
+/// [`DenseSlotMap`]: dense/struct.DenseSlotMap.html
 #[cfg(feature = "unstable")]
 pub trait Slottable {}
 
@@ -235,7 +259,7 @@ impl<T> Slottable for T {}
 ///
 /// [`Key`]: trait.Key.html
 /// [`BTreeMap`]: https://doc.rust-lang.org/std/collections/struct.BTreeMap.html
-#[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash)]
+#[derive(Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash)]
 pub struct KeyData {
     idx: u32,
     version: NonZeroU32,
@@ -284,6 +308,12 @@ impl KeyData {
     }
 }
 
+impl Debug for KeyData {
+    fn fmt(&self, f: &mut Formatter) -> fmt::Result {
+        write!(f, "{}v{}", self.idx, self.version.get())
+    }
+}
+
 impl Default for KeyData {
     fn default() -> Self {
         Self::null()
@@ -299,16 +329,9 @@ impl Default for KeyData {
 /// map. The easiest way to do this is through [`new_key_type!`], which
 /// makes a new type identical to [`DefaultKey`], just with a different name.
 ///
-/// This trait is intended to be a thin wrapper around [`KeyData`], and all
-/// methods must behave exactly as if we're operating on a [`KeyData`] directly.
-/// The internal unsafe code relies on this, therefore this trait is `unsafe` to
-/// implement. It is strongly suggested to simply use [`new_key_type!`] instead
-/// of implementing this trait yourself.
-///
 /// [`new_key_type!`]: macro.new_key_type.html
 /// [`DefaultKey`]: struct.DefaultKey.html
-/// [`KeyData`]: struct.KeyData.html
-pub unsafe trait Key: From<KeyData> + Into<KeyData> {
+pub trait Key: From<KeyData> + Into<KeyData> + Clone {
     /// Creates a new key that is always invalid and distinct from any non-null
     /// key. A null key can only be created through this method (or default
     /// initialization of keys made with [`new_key_type!`], which calls this
@@ -404,7 +427,7 @@ macro_rules! new_key_type {
             }
         }
 
-        unsafe impl $crate::Key for $name { }
+        impl $crate::Key for $name { }
 
         $crate::__serialize_key!($name);
 
